import { MessageTransmissionBase } from "./msg.transmission.base";
import { EventMessage, FisMessage, MessageTransmitter as MessageTransmitterInterface, TransmitterProfile } from '../interface/transport.interface'
import { ConnectionState, TransportEvent, TransportMessage } from "../interface/connector.interface";
import { v4 as uuidv4 } from 'uuid'
import { BehaviorSubject, distinct, distinctUntilChanged, filter, map, Observable, Subject } from "rxjs";
import { RetransmissionService } from "../utils/retransmission.service";
import { WrappedMessage } from "../utils/message.ordering";
import ConsoleLogger from "../utils/log.utils";
import { Adapter } from "../connector/adapter.base";
import { TransmitterAdapter } from "../connector/adapter.transmitter";

/* Take in all the messages that needs to be transported, and divide them accordingly. So the connector instances will do just that
connectors or adapters will have their own identifier*/
export class MessageTransmissionTransmitter extends MessageTransmissionBase implements MessageTransmitterInterface {
    private connectionStateEvent: BehaviorSubject<ConnectionState> = new BehaviorSubject<ConnectionState>('OFFLINE')
    private console: ConsoleLogger = new ConsoleLogger(`MessageTransmissionTransmitter`, ['transmission'])
    private messageToBeTransmitted!: Subject<FisMessage | WrappedMessage>
    transmitterProfile!: TransmitterProfile;
    retransmission!: RetransmissionService;

    constructor(profile: TransmitterProfile, adapter: TransmitterAdapter, event: Observable<TransportEvent>) {
        super()
        this.console.log({ message: `Constructing Transmitter Transmission with ${profile.name}` })
        this.event = event
        this.messageToBeTransmitted = new Subject()
        this.retransmission = new RetransmissionService()
        this.setTransmitter(profile)
        this.setUpAdapter(adapter)
        this.setupBuffer()

        // special case just for http in case of server/client disconnected, the unsent msg will be flushed back into messageToBeTransmitted
        this.uniqueHandlerToFlushUnsentMessages(event)
    }

    setupBuffer(): void {
        this.console.log({ message: `Setting up Retransmission Service...` })
        this.event.pipe(
            filter(event => (event.data as EventMessage).clientId == this.transmitterProfile.id),
            filter(event => event.event == 'Client Disconnected' || event.event == 'Client Re-connected' || event.event == 'Client Connected' || event.event == 'Server Disconnected' || event.event == 'Server Connected'),
            map(event => {
                if (event.event == 'Client Disconnected' || event.event == 'Server Disconnected') {
                    return 'OFFLINE'
                } else {
                    return `ONLINE`
                }
            }),
            distinctUntilChanged()
        ).subscribe((signal: ConnectionState) => {
            this.connectionStateEvent.next(signal)
            if(signal == 'OFFLINE') this.console.error({message: `${this.transmitterProfile.id} disconnected`})
            if(signal == 'ONLINE') this.console.log({message: `${this.transmitterProfile.id} connected`})
        })
        this.retransmission.implementRetransmission(this.messageToBeTransmitted, this.connectionStateEvent.asObservable(), true)
        // automatically subscribe to allow released bffered messages to be released
        this.retransmission.returnSubjectForBufferedItems().subscribe((message: WrappedMessage) => {
            // need to work with wrapped messages
            (this.mainAdapter as TransmitterAdapter).emit(message)
        })
    }

    setTransmitter(transmitterProfile: TransmitterProfile): void {
        this.transmitterProfile = transmitterProfile
    }

    emit(message: FisMessage): void {
        this.console.log({ message: `${this.connectionStateEvent.getValue() == 'ONLINE' ? `Transmitting ${message.header.messageID}` : `Buffering ${message.header.messageID}`}` })
        this.messageToBeTransmitted.next(message)
    }

    setUpAdapter(adapter: Adapter): void {
        // for now just hardcode to use 1 adapter type until connection manager is further enhacne to configure adapters on the fly
        this.mainAdapter = adapter
    }

    private uniqueHandlerToFlushUnsentMessages(event: Observable<TransportEvent>): void {
        event.pipe(
            filter(event => event.event == 'Re-Flush'),
            filter(event => (event.data as EventMessage).clientId == this.transmitterProfile.id),
        ).subscribe((event: TransportEvent) => {
            this.messageToBeTransmitted.next((((event.data as EventMessage).payload as TransportMessage).payload as WrappedMessage))
        })
    }

}